/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef ALIGNMENT_CPP
#define ALIGNMENT_CPP

#include "Alignment.h"

#include "AssemblyException.h"
#include <iostream>
#include <typeinfo>
#include <set>
using namespace std;


long Alignment::_count = 0;

Alignment::Alignment(){}

Alignment::Alignment(ScoredSeq * seqA, ScoredSeq* seqB, char orientation) :
  _locked(false),
  _seqA(seqA),
  _seqB(seqB),
  _orientation(orientation) {

  // -5 means "not paired yet"
  _aToB = new long[_seqA->size()];
  for (long n = 0; n < _seqA->size(); ++n){ _aToB[n] = -5; }
  _bToA = new long[_seqB->size()];
  for (long n = 0; n < _seqB->size(); ++n){ _bToA[n] = -5; }

  // -5 means "not gapped"
  _aGaps = new long[_seqA->size()];
  for (long n = 0; n < _seqA->size(); ++n){ _aGaps[n] = -5; }
  _bGaps = new long[_seqB->size()];
  for (long n = 0; n < _seqB->size(); ++n){ _bGaps[n] = -5; }

  //_count++;

  if (_orientation != '+' and _orientation != '-'){
    throw AssemblyException::ArgError("Alignment: invalid orientation char");
  }
  if (_seqA == _seqB){
    throw AssemblyException::ArgError("Alignment: two input ScoredSeqs cannot be the same");
  }
  _isNull = true; // a holder value
  //OK();
}


Alignment::Alignment(Alignment* oldAlignment, ScoredSeq* newSeqA, ScoredSeq* newSeqB) :
  _locked(false),
  _seqA(newSeqA),
  _seqB(newSeqB) {
  // copy all of the values from the old alignment
  if (! oldAlignment->_locked ){
    throw AssemblyException::ArgError("Alignment: you shouldn't be copying an unlocked alignment");
  }
  _orientation = oldAlignment->_orientation;
  _isNull = true; // a holder value
  // check the length requirement
  if ( _seqA->size() != oldAlignment->_seqA->size() ) {
    throw AssemblyException::ArgError("seqA sizes don't match in Alignment copying"); }
  if ( _seqB->size() != oldAlignment->_seqB->size() ) {
    cout << endl;
    cout << _seqB->size() << endl;
    cout << oldAlignment->_seqB->size() << endl;
    throw AssemblyException::ArgError("seqB sizes don't match in Alignment copying"); }
  // copy the arrays
  _aToB = new long[_seqA->size()];
  _aGaps = new long[_seqA->size()];
  for (long n = 0; n < _seqA->size(); ++n){
    _aToB[n] = oldAlignment->_aToB[n];
    _aGaps[n] = oldAlignment->_aGaps[n];
  }
  _bToA = new long[_seqB->size()];
  _bGaps = new long[_seqB->size()];
  for (long n = 0; n < _seqB->size(); ++n){
    _bToA[n] = oldAlignment->_bToA[n];
    _bGaps[n] = oldAlignment->_bGaps[n];
  }
  //_count++;
  lock();
  //OK();
}


Alignment::~Alignment(){
  ////OK();
  delete [] _aToB;
  delete [] _bToA;
  delete [] _aGaps;
  delete [] _bGaps;
  //_count--;
}



void Alignment::addLinkage(ScoredSeq * seq, long posSeq, long posPair){
  if ( _locked ){
    throw AssemblyException::ArgError("mutator method cannot be called because object is locked.");
  }

  if ( seq == _seqA ){
    if ( isLinked(posSeq, _seqA) or isLinked(posPair, _seqB) ){
      throw AssemblyException::ArgError("input position was already linked (1).");
    }
    if ( isGapped(posSeq, _seqA) ){
      throw AssemblyException::ArgError("in addLinkage, input position A was already gapped (1).");
    }
    if ( isGapped(posPair, _seqB) ){
      throw AssemblyException::ArgError("in addLinkage, input position B was already gapped (1).");
    }

    _aToB[posSeq] = posPair;
    _bToA[posPair] = posSeq;
  } else if ( seq == _seqB ){
    if ( isLinked(posSeq, _seqB) or isLinked(posPair, _seqA) ){
      throw AssemblyException::ArgError("input position was already linked (2).");
    }
    if ( isGapped(posSeq, _seqB) ){
      throw AssemblyException::ArgError("in addLinkage, input position B was already gapped (2).");
    }
    if ( isGapped(posPair, _seqA) ){
      throw AssemblyException::ArgError("in addLinkage, input position A was already gapped (2).");
    }
    _aToB[posPair] = posSeq;
    _bToA[posSeq] = posPair;
  } else {
    throw AssemblyException::ArgError("provided seq is not part of this alignment.");
  }
  //OK();
}

void Alignment::addLinkageBlock(ScoredSeq * seq, long posSeq, long posPair, long blockLength){
  for (long n = 0; n < blockLength; n++){
    addLinkage(seq, posSeq + n, posPair + n);
  }
  //OK();
}

void Alignment::addGap(ScoredSeq* inSeq, long posIns, long posGap){
  if ( _locked ){
    throw AssemblyException::ArgError("mutator method cannot be called because object is locked.");
  }
  if ( isLinked(posIns, inSeq) ){
    if ( inSeq == _seqA ){
      throw AssemblyException::ArgError("in addGap, input position A was already linked.");
    } else if ( inSeq == _seqB ){
      throw AssemblyException::ArgError("in addGap, input position B was already linked.");
    } else {
      throw AssemblyException::ArgError("provided seq is not part of this alignment.");
    }
  }
  if (posGap < -1){
    throw AssemblyException::ArgError("lowest gap pos after can be is -1.");
  }

  if ( inSeq == _seqA ){
    if (posGap >= _seqB->size()){
      throw AssemblyException::ArgError("highest gap pos after can be is seq length - 1.");
    }
    _aGaps[posIns] = posGap;
  } else if ( inSeq == _seqB ){
    if (posGap >= _seqA->size()){
      throw AssemblyException::ArgError("highest gap pos after can be is seq length - 1.");
    }
    _bGaps[posIns] = posGap;
  } else {
    throw AssemblyException::ArgError("provided seq is not part of this alignment.");
  }
  //OK();
}


void Alignment::addGapBlock(ScoredSeq* inSeq, long posIns, long posGap, long blockLength){
  //OK();
  for (long n = 0; n < blockLength; ++n){
    addGap(inSeq, posIns + n, posGap);
  }
  //OK();
}

ScoredSeq * Alignment::seqA(){
  //OK();
  return _seqA;
}

ScoredSeq* Alignment::seqB(){
  //OK();
  return _seqB;
}

char Alignment::orientation(){ 
  //OK();
  return _orientation;
}


bool Alignment::isNull(){ 
  if (! _locked ){
    throw AssemblyException::CallingError("cannot call 'isNull' if alignment is not locked.");
  }
  //OK();
  return _isNull;
}

bool Alignment::isLinked(long pos, ScoredSeq * seq){
  //OK();
  if (seq == _seqA) {
    if ( pos >= _seqA->size() ){
      cout << pos << " " << _seqA->size() << endl;
      throw AssemblyException::ArgError("Alignment::isLinked, pos was outside of seq");
    }
    return (_aToB[pos] != -5);
  }
  else if (seq == _seqB) {
    if ( pos >= _seqB->size() ){
      cout << pos << " " << _seqB->size() << endl;
      throw AssemblyException::ArgError("Alignment::isLinked, pos was outside of seq");
    }
    return (_bToA[pos] != -5);
  }
  else {
    throw AssemblyException::ArgError("provided seq is not part of this alignment.");
  }
}

bool Alignment::isGapped(long pos, ScoredSeq * seq){
  //OK();
  if (seq == _seqA) {
    if ( pos >= _seqA->size() ){
      cout << pos << " " << _seqA->size() << endl;
      throw AssemblyException::ArgError("Alignment::isGapped, pos was outside of seq");
    }
    return (_aGaps[pos] != -5);
  }
  else if (seq == _seqB) {
    if ( pos >= _seqB->size() ){
      cout << pos << " " << _seqB->size() << endl;
      throw AssemblyException::ArgError("Alignment::isGapped, pos was outside of seq");
    }
    return (_bGaps[pos] != -5);
  }
  else {
    throw AssemblyException::ArgError("provided seq is not part of this alignment.");
  }
}


long Alignment::getLinkage(long pos, ScoredSeq * seq){
  //OK();
  if (! isLinked(pos,seq) ){
    throw AssemblyException::ArgError("provided pos is not linked.");
  }
  if (seq == _seqA) { return _aToB[pos]; }
  else if (seq == _seqB) { return _bToA[pos]; }
  else { 
    throw AssemblyException::ArgError("provided seq is not part of this alignment.");
  }
}


long Alignment::gapPairedAfter(long pos, ScoredSeq * seq){
  if (! isGapped(pos,seq)){
    throw AssemblyException::ArgError("in alignment, cannot get gap pair for an ungapped position.");
  }
  //OK();
  if (seq == _seqA) { return _aGaps[pos]; }
  else if (seq == _seqB) { return _bGaps[pos]; }
  else { 
    throw AssemblyException::ArgError("provided seq is not part of this alignment.");
  }
}





long Alignment::score(AlignmentScoreMatrix* scoreMatrix, bool penalizeTerminalGaps){
  //OK();
  //  if (! _locked){
  if (! isLocked()){
    cout << int(_locked) << endl;
    throw AssemblyException::CallingError("cannot call 'calculateScore' if alignment is not locked.");
  }

  // the tally variable
  long scoreSum = 0;

  // establish where the beginning and end of each alignment are
  long posA = 0;
  long posB = 0;
  long sizeA = _seqA->size();
  long sizeB = _seqB->size();
  char senseA = '+';
  char senseB = _orientation;

  // make the decisions for the aligned segment
  while ( posA < sizeA or posB < sizeB ){

    if ( posA==sizeA or isGapped(posA,_seqA) or posB==sizeB or isGapped(posB,_seqB) ){
      long aStartPos = posA;    // unlike the combine method, here i start with
      long bStartPos = posB;    // the gapped position
      // bring both positions to the point after the alignment block
      while ( posA < sizeA and isGapped(posA,_seqA) ){ ++posA; }
      while ( posB < sizeB and isGapped(posB,_seqB) ){ ++posB; }

      // check that block is legit
      if ( posA!=sizeA and isLinked(posA,_seqA) and getLinkage(posA,_seqA)==posB ){} // this is legit
      else if ( posA==sizeA and posB==sizeB ){} // this is also legit
      else {
	throw AssemblyException::LogicError("in combine, conditions after a gap block are not legit.");
      }

      scoreSum += scoreGapBlock(aStartPos, posA, bStartPos, posB, scoreMatrix, penalizeTerminalGaps);
    } else { // MATCH
      scoreSum += scoreMatrix->match( _seqA->nucAtPosition(posA,senseA),
				      _seqB->nucAtPosition(posB,senseB) );
      posA++;
      posB++;
    }
  }
  return scoreSum;
}


// this is a helper for score
long Alignment::scoreGapBlock(long aStartPos, long aEndPos,
			     long bStartPos, long bEndPos,
			     AlignmentScoreMatrix* scoreMatrix,
			     bool penalizeTerminalGaps){

  long scoreTally = 0;

  // collect the sizes of the gaps that will be penalized by defining gap segments within the block
  map<long,long> aGapToSize;
  map<long,long> bGapToSize;
  getSegmentsFromBlock(&aGapToSize,aStartPos,aEndPos,_seqA,
		       _seqB->size()-1,penalizeTerminalGaps);
  getSegmentsFromBlock(&bGapToSize,bStartPos,bEndPos,_seqB,
		       _seqA->size()-1,penalizeTerminalGaps);

  for (map<long,long>::iterator it = aGapToSize.begin(); it != aGapToSize.end(); ++it){
    scoreTally += scoreMatrix->_newGap + (it->second - 1) * scoreMatrix->_extendGap;
  }
  for (map<long,long>::iterator it = bGapToSize.begin(); it != bGapToSize.end(); ++it){
    scoreTally += scoreMatrix->_newGap + (it->second - 1) * scoreMatrix->_extendGap;
  }
  return scoreTally;
}

// modifies "gapToSize"; this is a helper for scoreGapBlock
void Alignment::getSegmentsFromBlock(map<long,long>* gapToSize, long startPos, long endPos,
				     ScoredSeq* seq, long maxOpposingPosition,
				     bool penalizeTerminalGaps){
  long pos = startPos;
  long currentGapPos; // counts will be added to this gapToSize bin
  long priorGapAfter; // used to make sure that discontinuous gaps are separated

  while (pos < endPos){
    long currentGapAfter = gapPairedAfter(pos,seq);
    if ( penalizeTerminalGaps or (currentGapAfter!=-1 and currentGapAfter!=maxOpposingPosition) ){
      if ( pos==startPos or currentGapAfter!=priorGapAfter ){
	priorGapAfter = currentGapAfter;
	currentGapPos = pos;
	gapToSize->insert( pair<long,long> (currentGapPos, 1) );
      } else {
	map<long,long>::iterator currentGapPair = gapToSize->find( currentGapPos );
	if ( currentGapPair == gapToSize->end() ){
	  throw AssemblyException::LogicError("while scoring alignment, tried to add counts to a non-existant gap");
	} else {
	  currentGapPair->second ++;
	}
      }
    }
    ++pos;
  }
}





void Alignment::lock(){
  // figure out if the max position is a link in either case; if not, then
  // it is a gap, and the lower number of the two possibilities should be picked;
  // reset the min/max values
  long aMin = -2;
  long aMax = -2;
  long bMin = -2;
  long bMax = -2;

  // go through the tables and figure out the min and max coords that have been
  // assigned some sort of gap/pair assignment (terminal gaps will be added
  // outside of these coords).
  for (long aPos = 0; aPos < _seqA->size(); ++aPos){
    long bPos = _aToB[aPos];
    long aGap = _aGaps[aPos];
    if (bPos + aGap != -10){
      if ( aMin==-2 or aMin > aPos ){ aMin = aPos; }
      if ( aMax==-2 or aMax < aPos ){ aMax = aPos; }
      if (bPos != -5){
	if ( bMin==-2 or bMin > bPos ){ bMin = bPos; }
	if ( bMax==-2 or bMax < bPos ){ bMax = bPos; }
      }
    }

  }
  for (long bPos = 0; bPos < _seqB->size(); ++bPos){
    if (_bGaps[bPos] != -5){
      if ( bMin==-2 or bMin > bPos ){ bMin = bPos; }
      if ( bMax==-2 or bMax < bPos ){ bMax = bPos; }
    }
  }
  if ( aMin + aMax + bMin + bMax == -8 ){ _isNull = true; }
  else {
    _isNull = false;

    // take care of the side-by-side cases
    if ( aMin + aMax == -4 ){
      // one of these has to be >-2 because otherwise _isNull would be true
      if ( bMin == 0 and _bGaps[bMax] != -1 ){
	long newCoordA = _bGaps[0];
	addGap( _seqA, newCoordA, -1 );
	aMin = newCoordA;
	aMax = newCoordA; // its initial default value
      }
      if ( bMax == _seqB->size()-1 and _bGaps[bMax] != _seqA->size()-1 ){
	long newCoordA = _bGaps[bMax] + 1;
	addGap( _seqA, newCoordA, bMax );
	if (aMin > newCoordA){ aMin = newCoordA; }
	if (aMax < newCoordA){ aMax = newCoordA; }
      }
    } else if ( bMin + bMax == -4 ){
      // one of these has to be >-2 because otherwise _isNull would be true
      if ( aMin == 0 and _aGaps[aMax] != -1 ){
	long newCoordB = _aGaps[0];
	addGap( _seqB, newCoordB, -1 );
	bMin = newCoordB;
	bMax = newCoordB; // its initial default value
      }
      if ( aMax == _seqA->size()-1 and _aGaps[aMax] != _seqB->size()-1 ){
	long newCoordB = _aGaps[aMax] + 1;
	addGap( _seqB, newCoordB, aMax );
	if (bMin==-2 or bMin > newCoordB){ bMin = newCoordB; }
	if (bMin==-2 or bMax < newCoordB){ bMax = newCoordB; }
      }
    }

    // one min/max value should be at the edge of its sequence on each end
    if ( aMin != 0 and bMin != 0 ){
      throw AssemblyException::LogicError("neither sequence goes to the beginning");
    } else if ( aMax + 1 != _seqA->size() and bMax + 1 != _seqB->size() ){
      throw AssemblyException::LogicError("neither sequence goes to the end");
    }
    // fill in terminal gaps
    for (long n = 0; n < aMin; ++n){ addGap(_seqA, n, -1);  }
    for (long n = 0; n < bMin; ++n){ addGap(_seqB, n, -1);  }
    for (long n = aMax + 1; n < _seqA->size(); ++n){ addGap(_seqA, n, _seqB->size() - 1);  }
    for (long n = bMax + 1; n < _seqB->size(); ++n){ addGap(_seqB, n, _seqA->size() - 1);  }
  }
  // lock down and check rep
  _locked = true;
  //OK();
}


bool Alignment::isLocked(){
  //OK();
  return _locked;
}

void Alignment::OK(){
  if ( int(_locked)!=0 and int(_locked)!=1 ){
    cout << int(_locked) << endl;
    throw AssemblyException::LogicError("in Alignment, _locked is a nonsense value.");
  }
  if (_orientation!='+' and _orientation!='-'){
    cout << _orientation << endl;
    throw AssemblyException::LogicError("in Alignment, _orientation is a nonsense value.");
  }
  if ( _locked ){
    long aLinkCount = 0;
    long aGapCount = 0;
    for (long n = 0; n < _seqA->size(); ++n){
      if (_aToB[n] != -5){ ++aLinkCount; }
      if (_aGaps[n] != -5){ ++aGapCount; }
    }
    long bLinkCount = 0;
    long bGapCount = 0;
    for (long n = 0; n < _seqB->size(); ++n){
      if (_bToA[n] != -5){ ++bLinkCount; }
      if (_bGaps[n] != -5){ ++bGapCount; }
    }
    if ( aLinkCount != bLinkCount ){
      throw AssemblyException::LogicError("linkage maps are of unequal size.");
    }
    if (! _isNull ){
      if ( aLinkCount + aGapCount != _seqA->size() ){
	cout << endl;
	cout << aLinkCount << " " << aGapCount << " " << _seqA->size() << endl;
	cout << bLinkCount << " " << bGapCount << " " << _seqB->size() << endl;
	throw AssemblyException::LengthError("not all seqA positions are accounted for.");
      }
      if ( bLinkCount + bGapCount != _seqB->size() ){
	throw AssemblyException::LengthError("not all seqB positions are accounted for.");
      }
    }
    for (long aKey = 0; aKey < _seqA->size(); ++aKey){
      long bKey = _aToB[ aKey ];
      if ( bKey != -5 ){
	if ( _bToA[ bKey ] != aKey ){
	  throw AssemblyException::LogicError("linkage is not reciprocal");
	}
	if ( _aGaps[aKey] != -5 ){
	  throw AssemblyException::LogicError("linked position in A is also gapped");
	}
	if ( _bGaps[bKey] != -5 ){
	  throw AssemblyException::LogicError("linked position in B is also gapped");
	}
      }
    // check that everything flows in the right direction
    }
  }
}

#endif
