/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ECOFILTER_H
#define ECOFILTER_H

#include <set>
#include "ScoredSeq.h"
#include "ParameterizedInitialFile.h"
using namespace::std;

// this is a pure virtual class that will allow the contigs generated
// at the end of each cycle to be filtered according to some arbitrary
// criteria set by the implementing classes, which are below.

class EcoFilter {
 public:
  virtual ~EcoFilter();
  virtual EcoFilter* copy() = 0;
  virtual void filterSeqs(set<ScoredSeq*>* fullSet,
			  set<ScoredSeq*>* retainedSet,
			  set<ScoredSeq*>* removedSet,
			  int cycleNum) = 0;
};


// i am including the implementing classes in this file

// this one always returns true
class EcoFilterNull : public EcoFilter {
 public:
  EcoFilterNull();
  ~EcoFilterNull();
  EcoFilterNull* copy();
  void filterSeqs(set<ScoredSeq*>* fullSet,
		  set<ScoredSeq*>* retainedSet,
		  set<ScoredSeq*>* removedSet,
		  int cycleNum);
};

// makes sure a seq has enough score counts
class EcoFilterMinCount : public EcoFilter {
 public:
  EcoFilterMinCount();
  EcoFilterMinCount(float minCount);
  ~EcoFilterMinCount();
  EcoFilterMinCount* copy();
  void filterSeqs(set<ScoredSeq*>* fullSet,
		  set<ScoredSeq*>* retainedSet,
		  set<ScoredSeq*>* removedSet,
		  int cycleNum);
  static EcoFilterMinCount* defaultFilter();
 private:
  float _minCount;
};


// makes sure a seq is long enough
class EcoFilterMinLength : public EcoFilter {
 public:
  EcoFilterMinLength();
  EcoFilterMinLength(long minLength, int skipCycles);
  ~EcoFilterMinLength();
  EcoFilterMinLength* copy();
  void filterSeqs(set<ScoredSeq*>* fullSet,
		  set<ScoredSeq*>* retainedSet,
		  set<ScoredSeq*>* removedSet,
		  int cycleNum);
 private:
  long _minLength;
  int _skipCycles;
};


// seq must match among the initial contigs that have been included so far
class EcoFilterInitialContigMatch : public EcoFilter {
 public:
  EcoFilterInitialContigMatch();
  EcoFilterInitialContigMatch(float minFractId, int cyclesToSkip, bool fullFile = false);
  EcoFilterInitialContigMatch(float minFractId, int cyclesToSkip,
			      int numFilterCycles, int numSkipCycles, bool fullFile = false);
  ~EcoFilterInitialContigMatch();
  EcoFilterInitialContigMatch* copy();
  void filterSeqs(set<ScoredSeq*>* fullSet,
		  set<ScoredSeq*>* retainedSet,
		  set<ScoredSeq*>* removedSet,
		  int cycleNum);
  // class-specific
  void addFiles(set<ParameterizedInitialFile*>* initialContigFiles);
 private:
  set<ParameterizedInitialFile*> _initialContigFiles;
  float _minFractId;
  int _cyclesToSkip;
  int _numFilterCycles;
  int _numSkipCycles;
  bool _fullFile;
};

// imposes multiple tests
class EcoFilterMulti : public EcoFilter {
 public:
  EcoFilterMulti();
  EcoFilterMulti(EcoFilter** testArray, int numTests);
  ~EcoFilterMulti();
  EcoFilterMulti* copy();
  void filterSeqs(set<ScoredSeq*>* fullSet,
		  set<ScoredSeq*>* retainedSet,
		  set<ScoredSeq*>* removedSet,
		  int cycleNum);
 private:
  EcoFilter** _testArray;
  int _numTests;
};


#endif


